import {
  cloneElement,
  createElement,
  isValidElement,
  ReactNode,
  useEffect
} from "react";
import {
  AccessControlOptions,
  AccessFallbackComponent,
  CanParams
} from "./types";
import { useCan } from "./useCan";
import { useAccessControl } from "./useAccessControl";

type OnUnauthorizedProps = {
  reason?: string;
  params: CanParams;
};

export interface CanAccessProps extends CanParams {
  children: ReactNode;

  /**
   * Content to show if access control returns `false`
   */
  fallback?: AccessFallbackComponent;

  loading?: ReactNode

  queryOptions?: AccessControlOptions;

  /**
   * Callback function to be called if access control returns `can: false`
   */
  onUnauthorized?: (props: OnUnauthorizedProps) => void;
}

export function CanAccess(props: CanAccessProps) {
  const {
    children,
    fallback,
    loading,
    queryOptions,
    onUnauthorized,
    ...params
  } = props;

  const {
    isExecuting: isLoading,
    reason,
    can,
    error,
  } = useCan({
    ...params,
    queryOptions,
  });

  const {
    fallback: fallbackElement,
  } = useAccessControl({});

  useEffect(() => {
    if (onUnauthorized && can === false && !isLoading) {
      onUnauthorized({ reason, params });
    }
  }, [can, isLoading]);

  if (isLoading) {
    return loading;
  }

  if (can) {
    return children;
  }

  return resolveFallback(
    fallback ?? fallbackElement,
    reason,
    error,
  );
}

function resolveFallback(
  fallback: AccessFallbackComponent | undefined | null,
  reason: string | undefined,
  error: unknown,
) {
  if (fallback == null) {
    return null;
  }
  if (isValidElement(fallback)) {
    return cloneElement(fallback, { reason, error })
  }
  return createElement(fallback, { reason, error })
}
