import { ComponentType, ReactElement, ReactNode, useEffect, useState } from 'react';
import { Navigate, Route, Routes, To } from 'react-router-dom';
import { LogoutOnMount, useCheckAuth } from '../auth';
import { CoreLayoutProps } from './types';
import { getReactElement } from '../util';
import { DefaultLayout } from './DefaultLayout';
import { HasDashboardContextProvider } from './HasDashboardContextProvider';
import { useConfigureRoutesFromChildren } from './useConfigureRoutesFromChildren';
import { InitialLocationContextProvider } from '../routing';

export interface CoreAdminRoutesProps {
  /**
   * A catch-all react component to display when the URL does not match any
   *
   * @see https://marmelab.com/react-admin/Admin.html#catchall
   * @example
   * // in src/NotFound.js
   * import Card from '@mui/material/Card';
   * import CardContent from '@mui/material/CardContent';
   * import { Title } from 'react-admin';
   *
   * export const NotFound = () => (
   *     <Card>
   *         <Title title="Not Found" />
   *         <CardContent>
   *             <h1>404: Page not found</h1>
   *         </CardContent>
   *     </Card>
   * );
   *
   * // in src/App.js
   * import { Admin } from 'react-admin';
   * import { dataProvider } from './dataProvider';
   * import { NotFound } from './NotFound';
   *
   * const App = () => (
   *     <Admin catchAll={NotFound} dataProvider={dataProvider}>
   *         ...
   *     </Admin>
   * );
   */
  catchAll?: ComponentType<any> | ReactElement | null;

  children?: ReactNode;

  /**
   * @example
   * import { WithPermissions } from "rwas";
   * import HomepageView from "./views/Homepage";
   *
   * const authParams = {
   *     params: { route: 'homepage' },
   * };
   *
   * const firstpage = (
   *     <WithPermissions
   *         authParams={authParams}
   *         component={HomepageView}
   *     />
   * )
   */
  dashboard?: ComponentType | ReactElement | null;

  initialLocation?: To;

  /**
   * The main app layout component
   *
   * @see https://marmelab.com/react-admin/Admin.html#layout
   * @example
   * import { Admin, Layout } from 'react-admin';
   *
   * const MyLayout = ({ children }) => (
   *     <Layout appBarAlwaysOn>
   *        {children}
   *     </Layout>
   * );
   *
   * export const App = () => (
   *     <Admin dataProvider={dataProvider} layout={MyLayout}>
   *         ...
   *     </Admin>
   * );
   */
  layout?: ComponentType<CoreLayoutProps>;

  /**
   * The component displayed while fetching the auth provider if the admin child is an async function
   */
  loading?: ComponentType<any> | ReactElement | null;

  /**
   * The page to display when the admin has no Resource children
   *
   * @see https://marmelab.com/react-admin/Admin.html#ready
   * @example
   * import { Admin } from 'react-admin';
   *
   * const Ready = () => (
   *     <div>
   *         <h1>Admin ready</h1>
   *         <p>You can now add resources</p>
   *     </div>
   * )
   *
   * const App = () => (
   *     <Admin ready={Ready}>
   *         ...
   *     </Admin>
   * );
   */
  ready?: ComponentType<any> | ReactElement | null;

  /**
   * Flag to require authentication for all routes. Defaults to false.
   *
   * @see https://marmelab.com/react-admin/Admin.html#requireauth
   * @example
   * import { Admin } from 'react-admin';
   * import { dataProvider } from './dataProvider';
   * import { authProvider } from './authProvider';
   *
   * const App = () => (
   *     <Admin
   *         requireAuth
   *         authProvider={authProvider}
   *         dataProvider={dataProvider}
   *     >
   *         ...
   *     </Admin>
   * );
   */
  requireAuth?: boolean;
}

export function CoreAdminRoutes(props: CoreAdminRoutesProps) {
  const [routes, status] = useConfigureRoutesFromChildren(props.children);

  const {
    catchAll: catchAllElement,
    dashboard: dashboardElement,
    initialLocation,
    layout: Layout = DefaultLayout,
    loading: loadingElement,
    ready: readyElement,
    requireAuth = false,
  } = props;

  const [onlyAnonymousRoutes, setOnlyAnonymousRoutes] = useState(requireAuth);
  const [checkAuthLoading, setCheckAuthLoading] = useState(requireAuth);
  const checkAuth = useCheckAuth();

  useEffect(() => {
    if (requireAuth) {
      // do not log the user out on failure to allow access to custom routes with no layout
      // for other routes, the LogoutOnMount component will log the user out
      checkAuth(undefined, false)
        .then(() => {
          setOnlyAnonymousRoutes(false);
        })
        .catch(() => { })
        .finally(() => {
          setCheckAuthLoading(false);
        });
    }
  }, [checkAuth, requireAuth]);

  if (status === 'empty') {
    if (!readyElement) {
      throw new Error(
        'The app is empty. Please provide an empty component, ' +
        'or pass Resource or CustomRoutes as children.'
      );
    }
    return getReactElement(readyElement);
  }

  if (status === 'loading' || checkAuthLoading) {
    return (
      <Routes>
        <Route
          path="*"
          element={
            <div style={{ height: '100vh' }}>
              {loadingElement ? getReactElement(loadingElement) : 'loading...'}
            </div>
          }
        />
      </Routes>
    );
  }

  if (onlyAnonymousRoutes) {
    return (
      <Routes>
        <Route
          path="*"
          element={<LogoutOnMount />}
        />
      </Routes>
    );
  }

  return (
    <Routes>
      <Route
        path="/*"
        element={
          <HasDashboardContextProvider value={!!dashboardElement}>
            <InitialLocationContextProvider
              location={initialLocation}
              target="admin"
            >
              <Layout>
                <Routes>
                  {routes}
                  {dashboardElement ? (
                    <Route
                      index
                      element={
                        dashboardElement
                          ? (getReactElement(dashboardElement))
                          : initialLocation
                            ? (<Navigate to={initialLocation} />)
                            : null
                      }
                    />
                  ) : null}
                  {catchAllElement ? (
                    <Route
                      path="*"
                      element={getReactElement(catchAllElement)} />
                  ) : null}
                </Routes>
              </Layout>
            </InitialLocationContextProvider>
          </HasDashboardContextProvider>
        }
      />
    </Routes>
  );
}
