import {
  Children,
  Fragment,
  isValidElement,
  ReactElement,
  ReactNode,
  useCallback,
  useEffect,
  useState
} from "react";
import { Route as ReactRoute } from "react-router-dom";
import { useLogout, usePermissions } from "../auth";
import { useSafeSetState } from "../util";
import { CoreAdminContext, CoreAdminContextProps } from "./CoreAdminContext";
import { Route, RouteProps } from "../routing";

export type RenderRoutesFunction = (permissions: any) =>
  | ReactNode // (permissions) => <><Route /><Route /><Route /></>
  | Promise<ReactNode> // (permissions) => fetch().then(() => <><Route /><Route /><Route /></>)

export type AppChildren =
  | RenderRoutesFunction
  | Iterable<ReactNode | RenderRoutesFunction>
  | ReactNode;

export type AppRoutesStatus = 'loading' | 'empty' | 'ready';

function isRenderRoutesFunction(children: AppChildren): children is RenderRoutesFunction {
  return typeof children === "function";
}

function hasIteratorProtocol(children: AppChildren): children is Iterable<ReactNode | RenderRoutesFunction> {
  return (
    typeof Symbol !== "undefined" &&
    children != null &&
    typeof (children as any)[Symbol.iterator] === "function"
  );
}

function getRenderRoutesFunctions(children: AppChildren): RenderRoutesFunction[] {
  if (isRenderRoutesFunction(children)) {
    return [children];
  }
  if (Array.isArray(children)) {
    return children.reduce((acc, child) => {
      if (isRenderRoutesFunction(child)) {
        acc.push(child);
      }
      return acc;
    }, []);
  }
  if (hasIteratorProtocol(children)) {
    return getRenderRoutesFunctions(Array.from(children));
  }
  return [];
}

export function useConfigureRoutesFromChildren(children: AppChildren): [ReactElement[], AppRoutesStatus] {
  // Gather custom routes that were declared as direct children of AppRouter
  // e.g. Not returned from the child function (if any)
  // We need to know right away whether some resources were declared to correctly
  // initialize the status at the next stop
  const doLogout = useLogout();
  const { permissions, isPending } = usePermissions();
  const [routes, setRoutes] = useState(getRoutesFromNodes(children));
  const [status, setStatus] = useSafeSetState<AppRoutesStatus>(() => getStatus(children, routes));

  const mergeRoutes = useCallback((newRoutes: ReactElement[]) => {
    setRoutes(previous => previous.concat(newRoutes));
  }, [setRoutes]);

  if (!status) {
    throw new Error('Status should be defined');
  }

  useEffect(() => {
    let funcCounts = 0;

    const onResolve = () => {
      funcCounts -= 1;
      if (funcCounts <= 0) {
        setStatus('ready');
      }
    }

    const resolveChildFunction = async (childFunc: RenderRoutesFunction) => {
      try {
        const childrenFuncResult = childFunc(permissions);
        if ((childrenFuncResult as Promise<ReactNode>)?.then) {
          (childrenFuncResult as Promise<ReactNode>).then(
            resolvedChildren => {
              mergeRoutes(getRoutesFromNodes(resolvedChildren));
              onResolve();
            }
          );
        } else {
          mergeRoutes(getRoutesFromNodes(childrenFuncResult as ReactNode));
          onResolve();
        }
      } catch (error) {
        funcCounts -= 1;
        console.error(error);
        doLogout();
      }
    };

    const updateFromChildren = async () => {
      const functionChild = getRenderRoutesFunctions(children);
      const newRoutes = getRoutesFromNodes(children);
      mergeRoutes(newRoutes);
      setStatus(
        functionChild.length > 0
          ? 'loading'
          : newRoutes.length > 0
            ? 'ready'
            : 'empty'
      );

      if (functionChild.length) {
        funcCounts = functionChild.length;
        for (const fn of functionChild) {
          resolveChildFunction(fn);
        }
      } else {
        onResolve();
      }
    };

    if (!isPending) {
      updateFromChildren();
    }
  }, [
    children,
    doLogout,
    isPending,
    permissions,
    mergeRoutes,
    setRoutes,
    setStatus,
  ]);

  return [routes, status];
}

const getStatus = (
  children: AppChildren,
  routes: ReactNode[],
): AppRoutesStatus => {
  return getRenderRoutesFunctions(children).length > 0
    ? 'loading'
    : routes.length > 0
      ? 'ready'
      : 'empty';
};

/**
 * Inspect the children and return an array of routable elements
 */
function getRoutesFromNodes(children: AppChildren) {
  const routes: ReactElement[] = [];

  if (isRenderRoutesFunction(children)) {
    return routes;
  }

  if (
    !Array.isArray(children) &&
    !isValidElement(children) &&
    hasIteratorProtocol(children)
  ) {
    return getRoutesFromNodes(Array.from(children));
  }

  Children.forEach(children, node => {
    if (!isValidElement<any>(node)) {
      // Ignore non-elements. This allows people to more easily inline
      // conditionals in their route config.
      return;
    } else if (node.type === Fragment) {
      routes.push(...getRoutesFromNodes(node.props.children));
    } else if (node.type === CoreAdminContext) {
      // TODO ...
      const { basepath } = (node as ReactElement<CoreAdminContextProps>).props;
      routes.push(
        <ReactRoute
          key="adminContext"
          path={basepath}
          element={node}
        />
      )
    } else if (node.type === ReactRoute) {
      routes.push(node);
    } else if (node.type === Route) {
      const {
        name,
        remembeScrollPosition: _,
        authorised: __,
        children,
        ...props
      } = (node as ReactElement<RouteProps>).props;
      routes.push(
        // @ts-ignore
        <ReactRoute
          key={name}
          {...props}
          element={node}
          Component={null}
        >
          {getRoutesFromNodes(children)}
        </ReactRoute>
      );
    } else {
      // todo 与 react-router 报错保持一致
      // throw new Error("");
      // [div] is not a <Route> component. All component children of <Routes> must be a <Route> or <React.Fragment>
      throw new Error(
        "[" + node.type + "]  is not a <Route> component.  All component children of <Routes> must be a <Route> or <React.Fragment>"
      );
    }
  });

  return routes;
}
